/**
 * redpen: a text inspection tool
 * Copyright (c) 2014-2015 Recruit Technologies Co., Ltd. and contributors
 * (see CONTRIBUTORS.md)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package cc.redpen;

import cc.redpen.config.Configuration;
import cc.redpen.config.ValidatorConfiguration;
import cc.redpen.model.Document;
import cc.redpen.validator.ValidationError;
import junit.framework.Assert;
import org.junit.Before;
import org.junit.Test;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import static org.hamcrest.CoreMatchers.is;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertThat;
import static org.junit.matchers.JUnitMatchers.containsString;

public class RedPenTest {
    @Before
    public void setUp() {
        Locale.setDefault(Locale.ENGLISH);
    }

    @Test
    public void testEmptyValidator() throws RedPenException {

        List<Document> documents = new ArrayList<>();
        documents.add(new Document.DocumentBuilder()
                .addSection(1)
                .addParagraph()
                .addSentence("In a land far away, there once was as a hungry programmer.", 1)
                .addSentence("He was hungry for programming and programmed all day - "
                        + " - in Java, Python, C++, etc.", 2)
                .addSentence("Whe he wasn't programming, he was eating noodles.", 3)
                .addParagraph()
                .addSentence("One day while programming, he got a new idea.", 4)
                .build());

        Configuration configuration = new Configuration.ConfigurationBuilder().build();
        RedPen redPen = new RedPen(configuration);
        Map<Document, List<ValidationError>> errors = redPen.validate(documents);
        Assert.assertEquals(0, errors.get(documents.get(0)).size());
    }


    @Test
    public void testSentenceValidatorWithSimpleDocument()
            throws RedPenException {
        List<Document> documents = new ArrayList<>();
        documents.add(new Document.DocumentBuilder()
                .setFileName("tested file")
                .addSection(0)
                .addParagraph()
                .addSentence("it is a piece of a cake.", 0)
                .addSentence("that is also a piece of a cake.", 1)
                .build());

        RedPen redPen = getRedPenWithSentenceValidator();

        List<ValidationError> errors = redPen.validate(documents).get(documents.get(0));

        // validate the errors
        assertEquals(2, errors.size());
        for (ValidationError error : errors) {
            assertThat(error.getValidatorName(), is("SentenceLength"));
            assertThat(error.getMessage(),
                    containsString("exceeds the maximum "));
        }
    }

    @Test
    public void testSectionValidatorWithSimpleDocument()
            throws RedPenException {
        List<Document> documents = new ArrayList<>();

        documents.add(new Document.DocumentBuilder()
                .setFileName("tested file")
                .addSection(0)
                .addSectionHeader("foobar")
                .addParagraph()
                .addSentence("it is a piece of a cake.", 0)
                .addSentence("that is also a piece of a cake.", 1)
                .build());

        RedPen redPen = getRedPenWithSectionValidator();
        List<ValidationError> errors = redPen.validate(documents).get(documents.get(0));

        // validate the errors
        assertEquals(1, errors.size());
        for (ValidationError error : errors) {
            assertThat(error.getValidatorName(), is("SectionLength"));
            assertThat(error.getMessage(),
                    containsString("The number of the character in the section exceeds the maximum"));
        }
    }

    @Test
    public void testDocumentWithHeader() throws RedPenException {
        List<Document> documents = new ArrayList<>();
        documents.add(new Document.DocumentBuilder()
                .setFileName("tested file")
                .addSection(0)
                .addSectionHeader("this is it.")
                .addParagraph()
                .addSentence("it is a piece of a cake.", 0)
                .addSentence("that is also a piece of a cake.", 1)
                .build());

        RedPen redPen = getRedPenWithSentenceValidator();
        List<ValidationError> errors = redPen.validate(documents).get(documents.get(0));

        // validate the errors
        assertEquals(3, errors.size());
        for (ValidationError error : errors) {
            assertThat(error.getValidatorName(), is("SentenceLength"));
            assertThat(error.getMessage(),
                    containsString("exceeds the maximum "));
        }
    }

    @Test
    public void testDocumentWithList() throws RedPenException {
        List<Document> documents = new ArrayList<>();
        documents.add(new Document.DocumentBuilder()
                .setFileName("tested file")
                .addSection(0)
                .addSectionHeader("this is it")
                .addParagraph()
                .addSentence("it is a piece of a cake.", 0)
                .addSentence("that is also a piece of a cake.", 1)
                .addListBlock()
                .addListElement(0, "this is a list.")
                .build());

        RedPen redPen = getRedPenWithSentenceValidator();
        List<ValidationError> errors = redPen.validate(documents).get(documents.get(0));

        // validate the errors
        assertEquals(4, errors.size());
        for (ValidationError error : errors) {
            assertThat(error.getValidatorName(), is("SentenceLength"));
            assertThat(error.getMessage(),
                    containsString("exceeds the maximum "));
        }
    }

    @Test
    public void testDocumentWithoutContent() throws RedPenException {
        List<Document> documents = new ArrayList<>();
        documents.add(new Document.DocumentBuilder()
                .setFileName("tested file")
                .build());

        RedPen redPen = getRedPenWithSentenceValidator();
        List<ValidationError> errors = redPen.validate(documents).get(documents.get(0));

        // validate the errors
        assertEquals(0, errors.size());

    }

    private RedPen getRedPenWithSentenceValidator() throws
            RedPenException {

        Configuration configuration = new Configuration.ConfigurationBuilder()
                .addValidatorConfig(
                        new ValidatorConfiguration("SentenceLength").addAttribute("max_len", "5"))
                .build();
        return new RedPen(configuration);
    }

    private RedPen getRedPenWithSectionValidator() throws
            RedPenException {
        Configuration configuration = new Configuration.ConfigurationBuilder()
                .addValidatorConfig(
                        new ValidatorConfiguration("SectionLength").addAttribute("max_num", "5"))
                .build();
        return new RedPen(configuration);
    }
}
